from collections import defaultdict
import logging
import pathlib
import re

from ni_python_styleguide._acknowledge_existing_errors import _lint_errors_parser

EXCLUDED_ERRORS = {
    "BLK100",
}


class _InMultiLineStringChecker:
    def __init__(self, error_file):
        self._error_file = pathlib.Path(error_file)
        self._values = []
        self._load_lines()

    @property
    def values(self):
        return self._values

    def in_multiline_string(self, lineno):
        return self._values[lineno - 1]  # 0 indexed, but we number files 1 indexed

    @staticmethod
    def _count_multiline_string_endings_in_line(line):
        return line.count('"""') + line.count("'''")

    def _load_lines(self):
        in_file = self._error_file.read_text().splitlines()
        current_count = 0
        for line in in_file:
            line_value = (
                current_count
                + _InMultiLineStringChecker._count_multiline_string_endings_in_line(line)
            )
            # if occurances of multiline string markers is odd, this must be in a multiline
            self._values.append(line_value % 2 == 1)
            current_count = line_value


def _add_noqa_to_line(lineno, code_lines, error_code, explanation):
    line = code_lines[lineno]
    old_line_ending = "\n" if line.endswith("\n") else ""
    line = line.rstrip("\n")

    existing_suppression = re.search(r"noqa (?P<existing_suppresions>[\w\d]+\: [\w\W]+?) -", line)
    if existing_suppression:
        before = existing_suppression.groupdict()["existing_suppresions"]
        if error_code not in before:
            line = line.replace(before, before + f", {error_code}: {explanation}")
    else:
        line += f"  # noqa {error_code}: {explanation} (auto-generated noqa)"

    code_lines[lineno] = line + old_line_ending


def acknowledge_lint_errors(lint_errors):
    """Add a "noqa" comment for each of existing errors (unless excluded).

    Excluded error (reason):
    BLK100 - run black
    """
    parsed_errors = map(_lint_errors_parser.parse, lint_errors)
    parsed_errors = filter(None, parsed_errors)
    lint_errors_to_process = [error for error in parsed_errors if error not in EXCLUDED_ERRORS]

    lint_errors_by_file = defaultdict(list)
    for error in lint_errors_to_process:
        lint_errors_by_file[error.file].append(error)

    for bad_file, errors_in_file in lint_errors_by_file.items():
        path = pathlib.Path(bad_file)
        lines = path.read_text().splitlines(keepends=True)
        multiline_checker = _InMultiLineStringChecker(error_file=bad_file)

        # to avoid double marking a line with the same code, keep track of lines and codes
        handled_lines = defaultdict(list)
        for error in errors_in_file:
            skip = 0

            while multiline_checker.in_multiline_string(
                lineno=error.line + skip
            ) and error.line + skip < len(lines):
                # find when the multiline ends
                skip += 1

            cached_key = f"{error.file}:{error.line + skip}"
            if error.code in handled_lines[cached_key]:
                logging.warning(
                    "Multiple occurances of error %s code were logged for %s:%s, only suprressing first",
                    error.code,
                    error.file,
                    error.line + skip,
                )
                continue

            handled_lines[cached_key].append(error.code)

            _add_noqa_to_line(
                lineno=error.line - 1 + skip,
                code_lines=lines,
                error_code=error.code,
                explanation=error.explanation,
            )

        path.write_text("".join(lines))



import contextlib
from io import StringIO
import logging
import pathlib

import click
import flake8.main.application
import toml

from ni_python_styleguide import _acknowledge_existing_errors


def _qs_or_vs(verbosity):
    if verbosity != 0:
        return f"-{'v' * verbosity if verbosity > 0 else 'q' * abs(verbosity)}"
    return ""


def _read_pyproject_toml(ctx, param, value):
    value = value or "pyproject.toml"  # Only accept local pyproject.toml if not specified

    try:
        pyproject_data = toml.load(value)
    except FileNotFoundError:
        return None
    except (toml.TomlDecodeError, OSError) as e:
        raise click.FileError(filename=value, hint=f"Error reading configuration file: {e}")

    ctx.ensure_object(dict)
    ctx.obj["PYPROJECT"] = pyproject_data
    config = pyproject_data.get("tool", {}).get("ni-python-styleguide", {})

    config.pop("quiet", None)
    config.pop("verbose", None)

    if ctx.default_map is None:
        ctx.default_map = {}
    ctx.default_map.update(config)
    return value


def _get_application_import_names(pyproject):
    """Return the application package name the config."""
    # Otherwise override with what was specified
    app_name = (
        pyproject.get("tool", {})
        .get("ni-python-styleguide", {})
        .get("application-import-names", "")
    )

    # Allow the poetry name as a fallback
    if not app_name:
        app_name = pyproject.get("tool", {}).get("poetry", {}).get("name", "").replace("-", "_")

    return f"{app_name},tests"


class ConfigGroup(click.Group):
    """click.Group subclass which allows for a config option to load options from."""

    def __init__(self, *args, **kwargs):
        """Construct the click.Group with the config option."""
        kwargs["params"].append(
            click.Option(
                ["--config"],
                type=click.Path(
                    exists=True,
                    file_okay=True,
                    dir_okay=False,
                    readable=True,
                    allow_dash=False,
                    path_type=str,
                ),
                is_eager=True,
                callback=_read_pyproject_toml,
                help="Config file to load configurable options from",
            )
        )

        super().__init__(*args, **kwargs)


@click.group(cls=ConfigGroup)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Print more information. Repeat to increase verbosity.",
)
@click.option(
    "-q",
    "--quiet",
    count=True,
    help="Print less information. Repeat to decrease verbosity.",
)
@click.option(
    "--exclude",
    type=str,
    show_default=True,
    default="__pycache__,.git,.venv",
    help="Comma-separated list of files or directories to exclude.",
)
@click.option(
    "--extend-exclude",
    type=str,
    default="",
    help="Comma-separated list of files or directories to exclude (in addition to --exclude).",
)
@click.version_option()  # @TODO: override the message to include dependency version(s)
@click.pass_context
def main(ctx, verbose, quiet, config, exclude, extend_exclude):
    """NI's internal and external Python linter rules and plugins."""  # noqa: D4
    ctx.ensure_object(dict)
    ctx.obj["VERBOSITY"] = verbose - quiet
    ctx.obj["EXCLUDE"] = ",".join(filter(bool, [exclude.strip(","), extend_exclude.strip(",")]))
    ctx.obj["APP_IMPORT_NAMES"] = _get_application_import_names(ctx.obj.get("PYPROJECT", {}))


def _lint(obj, format, extend_ignore, file_or_dir):
    app = flake8.main.application.Application()
    args = [
        _qs_or_vs(obj["VERBOSITY"]),
        f"--config={(pathlib.Path(__file__).parent / 'config.ini').resolve()}",
        f"--exclude={obj['EXCLUDE']}" if obj["EXCLUDE"] else "",
        f"--format={format}" if format else "",
        f"--extend-ignore={extend_ignore}" if extend_ignore else "",
        # The only way to configure flake8-black's line length is through a pyproject.toml's
        # [tool.black] setting (which makes sense if you think about it)
        # So we need to give it one
        f"--black-config={(pathlib.Path(__file__).parent / 'config.toml').resolve()}",
        f"--application-import-names={obj['APP_IMPORT_NAMES']}",
        *file_or_dir,
    ]
    app.run(list(filter(bool, args)))
    app.exit()


@main.command()
# @TODO: When we're ready to encourage editor integration, add --diff flag
@click.option("--format", type=str, help="Format errors according to the chosen formatter.")
@click.option(
    "--extend-ignore",
    type=str,
    help="Comma-separated list of errors and warnings to ignore (or skip)",
)
@click.argument("file_or_dir", nargs=-1)
@click.pass_obj
def lint(obj, format, extend_ignore, file_or_dir):
    """Lint the file(s)/directory(s) given."""  # noqa: D4
    _lint(obj=obj, format=format, extend_ignore=extend_ignore, file_or_dir=file_or_dir)


@main.command()
@click.option(
    "--extend-ignore",
    type=str,
    help="Comma-separated list of errors and warnings to ignore (or skip)",
)
@click.argument("file_or_dir", nargs=-1)
@click.pass_obj
def acknowledge_existing_violations(obj, extend_ignore, file_or_dir):
    """Lint existing error and suppress.

    Use this command to acknowledge violations in existing code to allow for enforcing new code.
    """
    logging.info("linting code")
    capture = StringIO()
    with contextlib.redirect_stdout(capture):
        try:
            _lint(obj=obj, format=None, extend_ignore=extend_ignore, file_or_dir=file_or_dir)
        except SystemExit:
            pass  # the flake8 app wants to always SystemExit :(

    lines = capture.getvalue().splitlines()
    _acknowledge_existing_errors.acknowledge_lint_errors(lines)



from collections import namedtuple
import logging
import re

LintError = namedtuple("LintError", ["file", "line", "column", "code", "explanation"])


def parse(line):
    r"""
    Parse line into :class:`LintError`.

    >>> parse(r'source\arfile.py:55:16: BLK100 Black would make changes.')
    LintError(file='source\\arfile.py', line=55, column=16, code='BLK100', explanation='Black would make changes.')

    >>> parse(r"source\rpmfile\__init__.py:13:1: F401 'functools.wraps' imported but unused")
    LintError(file='source\\rpmfile\\__init__.py', line=13, column=1, code='F401', explanation="'functools.wraps' imported but unused")

    >>> parse(r"expected_output.py:77:6: N802 function name 'method_withBadName_with_bad_params_on_multiple_lines_1' should be lowercase")
    LintError(file='expected_output.py', line=77, column=6, code='N802', explanation="function name 'method_withBadName_with_bad_params_on_multiple_lines_1' should be lowercase")

    >>> parse(r"./tests/test_cli/acknowledge_existing_errors_test_cases__snapshots/doc_line_tests/expected_output.py:1:1: D100 Missing docstring in public module")
    LintError(file='./tests/test_cli/acknowledge_existing_errors_test_cases__snapshots/doc_line_tests/expected_output.py', line=1, column=1, code='D100', explanation='Missing docstring in public module')
    """  # NOQA W505: doc line too long (115 > 100 characters)
    p = Parser()
    return p.parse(line)


class Parser:
    """Lint errors parser."""

    __MATCHER = re.compile(
        r"^(?P<file>[\w\\/\.]+):(?P<line>\d+):(?P<column>\d+): (?P<code>\w+) (?P<explanation>.+)"
    )

    @staticmethod
    def _to_lint_error(file: str, line: str, column: str, code: str, explanation: str, **kwargs):
        return LintError(
            file=file,
            line=int(line),
            column=int(column),
            code=code,
            explanation=explanation,
            **kwargs
        )

    def parse(self, line):
        """Parse `line` and return a :class:`LintError`.

        :param line: the line to parse
        :return: lint error as metada object
        :rtype: LintError
        """
        data = Parser.__MATCHER.search(line)
        logging.debug("parsing line: %s, yielded %s", line, data)
        if not data:
            return None
        result = Parser._to_lint_error(**data.groupdict())
        return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ni_python_styleguide', 'ni_python_styleguide._acknowledge_existing_errors']

package_data = \
{'': ['*']}

install_requires = \
['black>=20.8b1,<21.0',
 'click>=7.1.2,<8.0.0',
 'flake8-black>=0.2.1,<0.3.0',
 'flake8-docstrings>=1.5.0,<2.0.0',
 'flake8-import-order>=0.18.1,<0.19.0',
 'flake8>=3.8.3,<4.0.0',
 'pep8-naming>=0.11.1,<0.12.0',
 'toml>=0.10.1,<0.11.0']

entry_points = \
{'console_scripts': ['ni-python-styleguide = ni_python_styleguide._cli:main']}

setup_kwargs = {
    'name': 'ni-python-styleguide',
    'version': '0.1.5',
    'description': "NI's internal and external Python linter rules and plugins",
    'long_description': "# NI Python Style Guide\n\n![logo](https://raw.githubusercontent.com/ni/python-styleguide/main/docs/logo.svg)\n\n---\n\n[![PyPI version](https://badge.fury.io/py/ni-python-styleguide.svg)](https://badge.fury.io/py/ni-python-styleguide) ![Publish Package](https://github.com/ni/python-styleguide/workflows/Publish%20Package/badge.svg) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n\nWelcome to NI's internal and external Python conventions and enforcement tooling.\n\n## Written Conventions\n\nOur written conventions can be found at https://ni.github.io/python-styleguide/.\n\nTheir source is in [docs/Coding-Conventions.md](https://github.com/ni/python-styleguide/tree/main/docs/Coding-Conventions.md).\n\nNOTE: Using the GitHub Pages link is preferable to a GitHub `/blob` link.\n\n## Enforcement tooling\n\nAs a tool, `ni-python-styleguide` is installed like any other script:\n\n```bash\npip install ni-python-styleguide\n```\n\n### Linting\n\nTo lint, just run the `lint` subcommand:\n\n```bash\nni-python-styleguide lint\n# or\nni-python-styleguide lint ./dir/\n# or\nni-python-styleguide lint module.py\n```\n\nThe rules enforced are all rules documented in the written convention, which are marked as enforced.\n\n### Formatting\n\n(This section to come!)\n\n### Editor Integration\n\n(This section to come!)\n",
    'author': 'NI',
    'author_email': 'opensource@ni.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ni/python-styleguide',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
